from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json

from compas.utilities import DataEncoder
from compas.utilities import DataDecoder

import compas
import os

import time
import inspect

from subprocess import Popen
from subprocess import PIPE

if compas.IPY:
    from .client_net import Client_Net as Client
    import Rhino
else:
    from .client_websockets import Client_Websokets as Client


__all__ = ['Proxy']

from functools import wraps
def retry_if_exception(ex, max_retries, wait = 0):
    def outer(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            assert max_retries > 0
            x = max_retries
            e = RuntimeError("unknown")
            while x:
                if compas.IPY:
                    Rhino.RhinoApp.Wait()
                try:
                    return func(*args, **kwargs)
                except ex as error:
                    e = error
                    print(e)
                    if isinstance(e, ServerSideError):
                        break
                    print('proxy call failed, trying time left:',x)
                    x -= 1
                    time.sleep(wait)
            raise e
        return wrapper
    return outer

class ServerSideError(Exception):
    pass

class Proxy():
    """Proxy is the interface between the user and a websocket client which communicates to websoket server in background.

    Parameters
    ----------
    port : int, optional
        The port number on the remote server.
        Default is ``9000``.

    Notes
    -----

    The service will make the correct (version of the requested) functionality available
    even if that functionality is part of a virtual environment. This is because it
    will use the specific python interpreter for which the functionality is installed to
    start the server.

    If possible, the proxy will try to reconnect to an already existing service

    The proxy will implement corresponding client with either python websokets library or
    .NET depending on environment.

    Examples
    --------

    .. code-block:: python

        from compas_cloud import Proxy
        p = Proxy()
        dr_numpy = p.package('compas.numerical.dr_numpy')


    """

    def __init__(self, host='127.0.0.1', port=9000, background=True, errorHandler=None):
        """init function that starts a remote server then assigns corresponding client(websockets/.net) to the proxy"""
        self._python = compas._os.select_python(None)
        self.host = host
        self.port = port
        self.background = background
        self.client = self.try_reconnect()
        if not self.client:
            self.client = self.start_server()
        self.callbacks = {}
        self.errorHandler = errorHandler

    def package(self, function, cache=False):
        raise RuntimeError("Proxy.package() has been deprecated, please use Proxy.function() instead.")
        

    def function(self, function, cache=False):
        """returns wrapper of function that will be executed on server side"""

        if self.errorHandler:
            @self.errorHandler
            @retry_if_exception(Exception, 5, wait = 0.5)
            def run_function(*args, **kwargs):
                return self.run(function, cache, *args, **kwargs)

            return run_function
        else:
            @retry_if_exception(Exception, 5, wait = 0.5)
            def run_function(*args, **kwargs):
                return self.run(function, cache, *args, **kwargs)
            
            return run_function

    def send(self, data):
        """encode given data before sending to remote server then parse returned result"""
        if not self.client:
            print("There is no connected client, try to restart proxy")
            return
        
        istring = json.dumps(data, cls=DataEncoder)
        self.client.send(istring)

        def listen_and_parse():
            result = self.client.receive()
            return json.loads(result, cls=DataDecoder)

        result = listen_and_parse()
        # keep receiving response until a non-callback result is returned
        while True:
            if isinstance(result, dict):
                if 'callback' in result:
                    cb = result['callback']
                    self.callbacks[cb['id']](*cb['args'], **cb['kwargs'])
                    result = listen_and_parse()
                elif 'listen' in result:
                    print(*result['listen'])
                    result = listen_and_parse()
                else:
                    break
            else:
                break

        return result

    def send_only(self, data):
        istring = json.dumps(data, cls=DataEncoder)
        return self.client.send(istring)

    def run(self, package, cache, *args, **kwargs):
        """pass the arguments to remote function and wait to receive the results"""
        args, kwargs = self.parse_callbacks(args, kwargs)
        idict = {'package': package, 'cache': cache,
                 'args': args, 'kwargs': kwargs}
        result = self.send(idict)
        if 'error' in result:
            raise ServerSideError("".join(result['error']))
        return result

    def Sessions(self, *args, **kwargs):
        return Sessions_client(self, *args, **kwargs)

    def get(self, cached_object):
        """get content of a cached object stored remotely"""
        idict = {'get': cached_object['cached']}
        return self.send(idict)

    def cache(self, data):
        """cache data or function to remote server and return a reference of it"""
        if callable(data):
            idict = {'cache_func': {
                'name': data.__name__,
                'source': inspect.getsource(data)
            }}
        else:
            idict = {'cache': data}
        return self.send(idict)

    def parse_callbacks(self, args, kwargs):
        """replace a callback functions with its cached reference then sending it to server"""
        for i, a in enumerate(args):
            cb = a
            if callable(cb):
                args[i] = {'callback': {'id': id(cb)}}
                self.callbacks[id(cb)] = cb
        for key in kwargs:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = {'callback': {'id': id(cb)}}
                self.callbacks[id(cb)] = cb
        return args, kwargs

    def try_reconnect(self):
        """try to reconnect to a existing server"""
        try:
            client = Client(self.host, self.port)
        except Exception:
            return None
        else:
            print("Reconnected to an existing server at {}:{}".format(self.host, self.port))
        return client

    def start_server(self):
        """use Popen to start a remote server in background"""
        env = compas._os.prepare_environment()

        args = [self._python, '-m', 'compas_cloud.server', str(self.port)]

        if self.background:
            print("Starting new cloud server in background at {}:{}".format(self.host, self.port))
            self._process = Popen(args, stdout=PIPE, stderr=PIPE, env=env)
        else:
            print("Starting new cloud server with prompt console at {}:{}".format(self.host, self.port))
            args[0] = compas._os.select_python('python')
            args = " ".join(args)
            os.system('start '+args)
        # import sys
        # self._process = Popen(args, stdout=sys.stdout, stderr=sys.stderr, env=env)

        success = False
        count = 20
        while count:
            if compas.IPY:
                    Rhino.RhinoApp.Wait()
            try:
                time.sleep(0.2)
                client = Client(self.host, self.port)
            except Exception as e:

                # stop trying if the subprocess is not running anymore
                if self.background:
                    if self._process.poll() is not None:
                        out, err =  self._process.communicate()
                        if out:
                            print(out.decode())
                        if err:
                            raise RuntimeError(err.decode())
                        raise RuntimeError('subprocess terminated, reason unknown')

                count -= 1
                print(e)
                print("\n\n    {} attempts left.".format(count))
            else:
                success = True
                break
        if not success:
            raise RuntimeError("The server is not available.")
        else:
            print("server started with port", self.port)

        return client

    def restart(self):
        """shut down and restart existing server and given ip and port"""
        self.client = self.try_reconnect()
        self.shutdown()
        time.sleep(1)
        self.client = self.start_server()

    def shutdown(self):
        """shut down currently connected server"""
        if self.client:
            if self.send_only({'control': 'shutdown'}):
                self.client = None
                print("server will shutdown and proxy client disconnected.")
        else:
            print("there is already no connected client")

    def check(self):
        """check if server connection is good"""
        return self.send({'control': 'check'})


class Sessions_client():

    def __init__(self, proxy, *args, **kwargs):
        self.proxy = proxy
        idict = {'sessions': {'command': 'create', 'args': args, 'kwargs': kwargs}}
        print(self.proxy.send(idict))

    def start(self):
        idict = {'sessions': {'command': 'start', 'args': (), 'kwargs': {}}}
        print(self.proxy.send(idict))

    def add_task(self, func, *args, **kwargs):
        cached = self.proxy.cache(func)
        idict = {'sessions': {'command': 'add_task', 'func': cached, 'args': args, 'kwargs': kwargs}}
        print(self.proxy.send(idict))

    def listen(self):
        idict = {'sessions': {'command': 'listen', 'args': (), 'kwargs': {}}}
        print(self.proxy.send(idict))

    def terminate(self):
        pass


if __name__ == "__main__":
    pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import compas

from .proxy import Proxy

if not compas.IPY:
    from .sessions import Sessions

__version__ = "0.1.1"

__all__ = [name for name in dir() if not name.startswith('_')]



import asyncio
import websockets

__all__ = ['Client_Websokets']


class Client_Websokets():
    """A Websoket client using webseckts and asyncio that works in a simple synchronous fashion

    Parameters
    ----------

    host: str, optional
        The host ip of remote server.
        Default is ``127.0.0.1``.

    port : int, optional
        The port number of remote server to connect to.
        Default is ``9000``.

    """

    def __init__(self, host='127.0.0.1', port=9000):
        """init the client, wait until it successfully connected to server"""
        async def connect():
            uri = "ws://{}:{}".format(host, str(port))
            self.websocket = await websockets.connect(uri, max_size=2**30)
        asyncio.get_event_loop().run_until_complete(connect())
        print('connected to cloud using websockets client!')

    def send(self, payload):
        """send a message to server and wait until sent"""
        async def _send():
            await self.websocket.send(payload)
            return True
        return asyncio.get_event_loop().run_until_complete(_send())

    def receive(self):
        """listen to a message until received one"""
        async def _receive():
            return await self.websocket.recv()
        return asyncio.get_event_loop().run_until_complete(_receive())



from autobahn.asyncio.websocket import WebSocketServerProtocol

from compas.utilities import DataEncoder
from compas.utilities import DataDecoder
import importlib
import json
from compas_cloud import Sessions
from threading import Thread
from multiprocessing import Queue
import time
import sys
import traceback


class CompasServerProtocol(WebSocketServerProtocol):
    """The CompasServerProtocol defines the behaviour of compas cloud server"""
    cached = {}
    sessions = None

    def onConnect(self, request):
        """print client info on connection"""
        print("Client connecting: {}".format(request.peer))

    def onClose(self, wasClean, code, reason):
        """print reason on connection closes"""
        print("WebSocket connection closed: {}".format(reason))

    def onMessage(self, payload, isBinary):
        """process the income messages"""
        result = self.process(payload)
        self.sendMessage(result.encode(), isBinary)

    def callback(self, _id, *args, **kwargs):
        """send the arguments of callback functions to client side"""
        data = {'callback': {'id': _id, 'args': args, 'kwargs': kwargs}}
        istring = json.dumps(data, cls=DataEncoder)
        self.sendMessage(istring.encode())

    def load_cached(self, data):
        """detect and load cached data or callback functions in arguments"""
        for i, a in enumerate(data['args']):
            if isinstance(a, dict):
                if 'cached' in a:
                    data['args'][i] = self.cached[a['cached']]

        for key in data['kwargs']:
            if isinstance(data['kwargs'][key], dict):
                if 'cached' in data['kwargs'][key]:
                    data['kwargs'][key] = self.cached[data['kwargs'][key]['cached']]
                if 'callback' in data['kwargs'][key]:
                    _id = data['kwargs'][key]['callback']['id']
                    self.cached[_id] = lambda *args, **kwargs: self.callback(
                        _id, *args, **kwargs)
                    data['kwargs'][key] = self.cached[_id]

    def execute(self, data):
        """execute corresponding binded functions with received arguments"""
        package = data['package']
        names = package.split('.')
        name = '.'.join(names[:-1])
        module = importlib.import_module(name)
        function = getattr(module, names[-1])

        start = time.time()
        print('running:', package)
        self.load_cached(data)

        if data['cache']:
            to_cache = function(*data['args'], **data['kwargs'])
            self.cached[id(to_cache)] = to_cache
            result = {'cached': id(to_cache)}
        else:
            result = function(*data['args'], **data['kwargs'])
        t = time.time()-start
        print('finished in: {}s'.format(t))
        return result

    def get(self, data):
        """get cached data from its id"""
        _id = data['get']
        return self.cached[_id]

    def cache(self, data):
        """cache received data and return its reference object"""
        to_cache = data['cache']
        _id = id(to_cache)
        self.cached[_id] = to_cache
        return {'cached': _id}

    def cache_func(self, data):
        """cache a excutable function"""
        name = data['cache_func']['name']
        exec(data['cache_func']['source'])
        exec('self.cached[name] = {}'.format(name))
        return {'cached_func': name}

    def sessions_alive(self):
        return isinstance(self.sessions, Sessions)

    def control(self, data):
        command = data['control']
        if command == 'shutdown':
            raise KeyboardInterrupt
        if command == 'check':
            print('check from client')
            return {'status': "I'm good"}
            
        raise ValueError("Unrecognised control command")

    def control_sessions(self, data):
        """control attached sessions according to message received"""
        s = data["sessions"]
        if s["command"] == 'create':
            if not self.sessions_alive():
                self.sessions = Sessions(socket=self)
                return "session successfully created"
            else:
                raise RuntimeError("There is already sessions running, try to reconnect or shut down")
        else:
            if not self.sessions_alive():
                raise RuntimeError("There no running sessions, try to create one first")

            if s["command"] == 'add_task':
                func_id = s['func']['cached_func']
                func = self.cached[func_id]
                self.sessions.add_task(func, *s['args'], *s['kwargs'])
                return "task added"

            if s["command"] == 'start':
                self.sessions.start()
                return "sessions started"

            if s["command"] == 'listen':
                self.sessions.listen()
                self.sessions = None
                return "All sessions concluded"

            if s["command"] == 'shutdown':
                self.sessions.terminate()
                self.sessions = None

    def process(self, data):
        """process received data according to its content"""
        data = json.loads(data, cls=DataDecoder)

        try:

            if 'cache' in data:
                result = self.cache(data)

            if 'cache_func' in data:
                result = self.cache_func(data)

            if 'package' in data:
                result = self.execute(data)

            if 'get' in data:
                result = self.get(data)

            if 'sessions' in data:
                result = self.control_sessions(data)

            if 'control' in data:
                result = self.control(data)

        except BaseException as error:

            if isinstance(error, KeyboardInterrupt):
                raise KeyboardInterrupt
            
            exc_type, exc_value, exc_tb = sys.exc_info()
            result = {'error': traceback.format_exception(exc_type, exc_value, exc_tb)}
            print("".join(result['error']))

        istring = json.dumps(result, cls=DataEncoder)
        return istring


if __name__ == '__main__':

    try:
        import asyncio
    except ImportError:
        # Trollius >= 0.3 was renamed
        import trollius as asyncio

    from autobahn.asyncio.websocket import WebSocketServerFactory
    factory = WebSocketServerFactory()
    factory.protocol = CompasServerProtocol

    loop = asyncio.get_event_loop()
    coro = loop.create_server(factory, '127.0.0.1', 9000)
    server = loop.run_until_complete(coro)
    print("starting compas_cloud server")

    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    finally:
        print("shuting down server")
        server.close()
        loop.close()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import math

from System import Array
from System import ArraySegment
from System import Byte
from System import Uri
from System import UriBuilder
from System.Net.WebSockets import ClientWebSocket
from System.Net.WebSockets import WebSocketCloseStatus
from System.Net.WebSockets import WebSocketMessageType
from System.Net.WebSockets import WebSocketReceiveResult
from System.Net.WebSockets import WebSocketState
from System.Text import Encoding
from System.Threading import CancellationTokenSource
from System.Threading import SemaphoreSlim

import time

SEND_CHUNK_SIZE = 1024
RECEIVE_CHUNK_SIZE = 1024


__all__ = ['Client_Net']


class Client_Net():
    """A Websoket client using .NET that works in a simple synchronous fashion

    Parameters
    ----------

    host: str, optional
        The host ip of remote server.
        Default is ``127.0.0.1``.

    port : int, optional
        The port number of remote server to connect to.
        Default is ``9000``.

    """
    def __init__(self, host='127.0.0.1', port=9000):
        """init the client, wait until it successfully connected to server"""
        scheme = 'ws'
        builder = UriBuilder(scheme, host, port)
        uri = builder.Uri

        self.token = CancellationTokenSource().Token
        self.socket = ClientWebSocket()
        task = self.socket.ConnectAsync(uri, self.token)
        task.Wait()
        print('connected to cloud using .NET client!')

    def disconnect(self):
        """disconnect from server"""
        task = self.socket.CloseAsync(
            WebSocketCloseStatus.NormalClosure, 'script finished', self.token)
        task.Wait()
        print('closed!')

    def send(self, payload):
        """send a message to server and wait until sent"""
        if self.socket.State != WebSocketState.Open:
            raise RuntimeError('Connection is not open.')

        message_buffer = Encoding.UTF8.GetBytes(payload)
        message_length = len(message_buffer)
        chunks_count = int(math.ceil(float(message_length) / SEND_CHUNK_SIZE))
        i = 0

        while True:
            # print('sending chunk', i)
            offset = SEND_CHUNK_SIZE * i
            is_last_message = (i == chunks_count - 1)

            if is_last_message:
                count = message_length - offset
            else:
                count = SEND_CHUNK_SIZE
            message_chunk = ArraySegment[Byte](message_buffer, offset, count)
            task = self.socket.SendAsync(
                message_chunk, WebSocketMessageType.Text, is_last_message, self.token)
            task.Wait()
            i += 1

            if is_last_message:
                return True

    def receive(self):
        """listen to a message until received one"""
        if self.socket.State != WebSocketState.Open:
            raise RuntimeError('Connection is not open.')

        chunks = []
        while True:
            # print('receive chunk')
            buffer = Array.CreateInstance(Byte, RECEIVE_CHUNK_SIZE)
            task = self.socket.ReceiveAsync(
                ArraySegment[Byte](buffer), self.token)
            task.Wait()
            chunk = Encoding.UTF8.GetString(buffer)
            chunks.append(chunk.rstrip('\x00'))
            if task.Result.EndOfMessage:
                return ''.join(chunks)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
import os
import sys
import json
from multiprocessing import Process, Queue, cpu_count
from contextlib import contextmanager
import traceback
from threading import Thread

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

class CapturedText(object):
    log_path = None


@contextmanager
def captured(name=None, log_path=None):

    stdout = sys.stdout
    stderr = sys.stderr


    if log_path:
        sys.stdout = sys.stderr = open(log_path, "w", 1)
    else:
        sys.stdout = sys.stderr = StringIO()
    c = CapturedText()
    c.name = name
    c.outfile = sys.stdout
    c.log_path = log_path

    yield c

    if log_path:
        c.outfile.close()

    sys.stdout = stdout
    sys.stderr = stderr



TASK_FINISHED = "____FINISHED____"


class Sessions():
    """a task-manager class that helps to execute a batch of long-lasting tasks such as FEA and DEM simulations.

    Parameters
    ----------
    log_path : str, optional
        The folder path to store individual task logs.
        Default is ``None``.
    worker_num : int, optional
        The number of workers to execute tasks in parallel.
        Default is equal to number of available CPUs.
    socket: internal use only

    Examples
    --------

    .. code-block:: python

        from compas_cloud import Sessions

        def func(a):
            import time
            for i in range(a):
                time.sleep(1)
                print('sleeped ', i, 's')

        s.add_task(func, 1)
        s.start()
        s.listen()

    """
    def __init__(self, log_path=None, worker_num=None, socket=None):
        """init function"""
        self.counter = 0
        self.tasks = {}
        self.waiting = Queue()
        self.messages = Queue()
        self.log_path = log_path
        self.worker_num = worker_num
        self.socket = socket

    def add_task(self, func, *args, **kwargs):
        """add a task function and its input parameters to the queue"""
        task = {"func": func, "args": args, "kwargs": kwargs, "status": "waiting"}
        _id = len(self.tasks)
        if self.log_path is not None:
            task["log_path"] = os.path.join(self.log_path, "task-{}.log".format(_id))
        else:
            task["log_path"] = None
        self.tasks[_id] = task
        self.waiting.put(_id)

    def create_workers(self, worker_num=None):

        def worker(waiting, messages, tasks):
            pid = os.getpid()
            messages.put(("message", "worker {} started".format(pid)))
            while not waiting.empty():
                task_id = waiting.get()
                task = tasks[task_id]
                messages.put(("task_running", task_id))
                with captured(task_id, log_path=task["log_path"]) as c:
                    def output_reader(proc):
                        if proc.log_path:
                            messages.put(("message", "task-{}: streaming log to {}".format(proc.name, proc.log_path)))
                            return
                        out = proc.outfile
                        lastpos = 0
                        while True:
                            if out.tell() != lastpos:
                                out.seek(lastpos)
                                line = out.read()
                                if line[-16:] == TASK_FINISHED:
                                    break
                                messages.put(("task_log", "task-{} log: {}".format(proc.name, line)))
                                lastpos = out.tell()
                            time.sleep(0.05)

                    t = Thread(target=output_reader, args=(c,))
                    t.start()

                    try:
                        task["func"](*task["args"], **task["kwargs"])
                        print(TASK_FINISHED, end="")
                        t.join()
                        messages.put(("task_finished", task_id))
                        c.finished = True
                    except Exception:
                        traceback.print_exc()
                        print(TASK_FINISHED, end="")
                        t.join()
                        messages.put(("task_failed", task_id))

            messages.put(("message", "worker {} terminated".format(pid)))

        if self.worker_num is None:
            self.worker_num = cpu_count()
        if self.worker_num > len(self.tasks):
            self.worker_num = len(self.tasks)

        self.log("using {} workers".format(self.worker_num))
        self.workers = [Process(target=worker, args=(self.waiting, self.messages, self.tasks)) for i in range(self.worker_num)]

    def process_message(self):

        msg_type, content = self.messages.get()

        if msg_type == "task_running":
            key = content
            self.tasks[key]["status"] = "running"
            self.log("task-{}: started".format(key))

        elif msg_type == "task_finished":
            key = content
            self.tasks[key]["status"] = "finished"
            self.log("task-{}: finished".format(key))

        elif msg_type == "task_failed":
            key = content
            self.tasks[key]["status"] = "failed"
            self.log("task-{}: failed".format(key))
        elif msg_type == "task_log":
            self.log(content, end="")
        else:
            self.log(content)

    def log(self, *args, **kwargs):
        print(self.status, "________", *args, **kwargs)
        if self.socket is not None:
            msg = (self.status, "________", args)
            data = json.dumps({"listen": msg})
            self.socket.sendMessage(data.encode())

    def start(self):
        """kick off the execution of tasks"""
        self.log("START")
        self.create_workers()
        for worker in self.workers:
            worker.start()

    def listen(self):
        """listen to the task messages until all finished"""
        while not self.all_finished() or not self.messages.empty():
            self.process_message()
        self.log("FINISHED")

    @property
    def status(self):
        s = {"waiting": 0, "running": 0, "failed": 0, "finished": 0, "total": len(self.tasks)}
        for k in self.tasks:
            for key in s:
                if self.tasks[k]['status'] == key:
                    s[key] += 1
        return s

    def all_finished(self):
        return self.status["finished"] + self.status["failed"] == self.status["total"]

    def terminate(self):
        for w in self.workers:
            w.terminate()

    def summary(self):
        pass


if __name__ == '__main__':

    def func(a):
        for i in range(a):
            time.sleep(1)
            print('sleeped ', i, 's')

        # raise RuntimeError('error example')
        return a

    s = Sessions(log_path="temp", worker_num=4)

    s.add_task(func, 1)
    s.add_task(func, 2)
    s.add_task(func, 3)
    s.add_task(func, 4)
    s.add_task(func, 5)


    s.start()
    s.listen()



import asyncio
import websockets
import time

from compas.utilities import DataEncoder
from compas.utilities import DataDecoder
import importlib
import json
from compas_cloud import Sessions
import functools

from multiprocessing import Queue, Process

class Sessions_server(Sessions):

    socket_message = Queue()

    def log(self, *args, **kwargs):
        print(self.status, "________", *args, **kwargs)
        self.socket_message.put((self.status, "________", args))

    async def listen(self, socket):
        while not self.all_finished() or not self.messages.empty():
            self.process_message()
            await self.send_to_socket(socket)
        self.log("FINISHED")
        await self.send_to_socket(socket)

    async def send_to_socket(self, socket):
        while not self.socket_message.empty():
            msg = self.socket_message.get()
            data = json.dumps({"listen": msg})
            await socket.send(data)



class Server_Websokets():

    def __init__(self, host='127.0.0.1', port=9000):
        """init the client, wait until it successfully connected to server"""

        self.cached = {}
        self.websocket = None

        async def user_session(websocket, path):
            self.websocket = websocket
            print('user connected', websocket, path)
            self.messages_to_send = Queue()

            async def listen():
                while True:
                    try:
                        data = await self.websocket.recv()
                        result = await self.process(data)
                        await self.websocket.send(result)

                    except Exception as e:
                        print('user disconnected:', e)
                        break

            async def wait():
                i = 0
                while i < 20:

                    if self.messages_to_send.empty():
                        # print('empty')
                        pass
                    else:
                        i += 1
                        # print(self.messages_to_send.get())
                        print('callback', i)
                        # print(self.messages_to_send.get())
                        await self.websocket.send(self.messages_to_send.get())

                    await asyncio.sleep(0.01)

            await asyncio.wait([listen(), wait()])

            self.websocket = None

        start_server = websockets.serve(user_session, host, port)
        self.loop = asyncio.get_event_loop()
        self.loop.run_until_complete(start_server)
        print('started server')
        self.loop.run_forever()



    def callback(self, _id, *args, **kwargs):
        """send the arguments of callback functions to client side"""
        data = {'callback': {'id': _id, 'args': args, 'kwargs': kwargs}}
        istring = json.dumps(data, cls=DataEncoder)
        self.messages_to_send.put(istring)



    def load_cached(self, data):
        """detect and load cached data or callback functions in arguments"""
        for i, a in enumerate(data['args']):
            if isinstance(a, dict):
                if 'cached' in a:
                    data['args'][i] = self.cached[a['cached']]

        for key in data['kwargs']:
            if isinstance(data['kwargs'][key], dict):
                if 'cached' in data['kwargs'][key]:
                    data['kwargs'][key] = self.cached[data['kwargs'][key]['cached']]
                if 'callback' in data['kwargs'][key]:
                    _id = data['kwargs'][key]['callback']['id']
                    self.cached[_id] = lambda *args, **kwargs: self.callback(
                        _id, *args, **kwargs)
                    data['kwargs'][key] = self.cached[_id]

    def execute(self, data):
        """execute corresponding binded functions with received arguments"""
        package = data['package']
        names = package.split('.')
        name = '.'.join(names[:-1])
        module = importlib.import_module(name)
        function = getattr(module, names[-1])

        self.load_cached(data)

        if data['cache']:
            to_cache = function(*data['args'], **data['kwargs'])
            self.cached[id(to_cache)] = to_cache
            result = {'cached': id(to_cache)}
        else:
            print('start execution of function')
            result = function(*data['args'], **data['kwargs'])

        print('execution finished')
        return result

    def get(self, data):
        """get cached data from its id"""
        _id = data['get']
        return self.cached[_id]

    def cache(self, data):
        """cache received data and return its reference object"""
        to_cache = data['cache']
        _id = id(to_cache)
        self.cached[_id] = to_cache
        return {'cached': _id}

    def cache_func(self, data):
        """cache a excutable function"""
        name = data['cache_func']['name']
        exec(data['cache_func']['source'])
        exec('self.cached[name] = {}'.format(name))
        return {'cached_func': name}

    def sessions_alive(self):
        if hasattr(self, 'sessions'):
            if isinstance(self.sessions, Sessions):
                return True
        return False

    async def control_sessions(self, data):
        s = data["sessions"]
        if s["command"] == 'create':
            if not self.sessions_alive():
                self.sessions = Sessions_server()
                return "session successfully created"
            else:
                raise RuntimeError("There is already sessions running, try to reconnect or shut down")
        else:
            if not self.sessions_alive():
                raise RuntimeError("There no running sessions, try to create one first")

            if s["command"] == 'add_task':
                func_id = s['func']['cached_func']
                func = self.cached[func_id]
                self.sessions.add_task(func, *s['args'], *s['kwargs'])
                return "task added"

            if s["command"] == 'start':
                self.sessions.start()
                return "sessions started"

            if s["command"] == 'listen':
                await self.sessions.listen(self.websocket)
                # import time
                # for i in range(10):
                #     time.sleep(1)
                #     print('sending to socket')
                #     istring = json.dumps({"listen": ("listen", i)})
                #     # print(self.sendMessage(istring.encode()))
                #     await self.websocket.send(istring)

                return "Session concluded"

            if s["command"] == 'shutdown':
                self.sessions.terminate()
                self.sessions = None

    async def process(self, data):
        """process received data according to its content"""
        data = json.loads(data, cls=DataDecoder)

        try:

            if 'cache' in data:
                result = self.cache(data)

            if 'cache_func' in data:
                result = self.cache_func(data)

            if 'package' in data:
                result = self.execute(data)

            if 'get' in data:
                result = self.get(data)

            if 'sessions' in data:
                result = await self.control_sessions(data)

        except BaseException as error:
            result = {'error': '{}:{}'.format(type(error).__name__, error)}
            print(result)

        istring = json.dumps(result, cls=DataEncoder)
        return istring

# if __name__ == "main":
Server_Websokets()
